"""
Multiple Hypothesis Tracking (MHT) implementation.

MHT maintains multiple hypotheses about measurement-to-track associations,
deferring hard decisions until more information is available. This allows
the tracker to recover from association errors.

This implementation uses track-oriented MHT with N-scan pruning.

References
----------
.. [1] S. Blackman and R. Popoli, "Design and Analysis of Modern
       Tracking Systems," Artech House, 1999.
.. [2] D. Reid, "An Algorithm for Tracking Multiple Targets,"
       IEEE Trans. Automatic Control, 1979.
"""

from typing import Callable, Dict, List, NamedTuple, Optional, Set

import numpy as np
from numpy.typing import ArrayLike, NDArray
from scipy.stats import chi2

from pytcl.assignment_algorithms.gating import mahalanobis_distance
from pytcl.trackers.hypothesis import (
    Hypothesis,
    HypothesisTree,
    MHTTrack,
    MHTTrackStatus,
    generate_joint_associations,
)


class MHTConfig(NamedTuple):
    """Configuration for MHT tracker.

    Attributes
    ----------
    n_scan : int
        Number of scans for N-scan pruning. Default 3.
    max_hypotheses : int
        Maximum number of hypotheses to maintain. Default 100.
    detection_prob : float
        Probability of detection (Pd). Default 0.9.
    clutter_density : float
        Spatial density of false alarms. Default 1e-6.
    gate_probability : float
        Gating probability for chi-squared test. Default 0.99.
    confirm_threshold : int
        Number of hits to confirm a track. Default 3.
    delete_threshold : int
        Number of consecutive misses to delete a track. Default 5.
    min_hypothesis_prob : float
        Minimum hypothesis probability. Default 1e-6.
    new_track_weight : float
        Prior weight for new track hypothesis. Default 0.1.
    """

    n_scan: int = 3
    max_hypotheses: int = 100
    detection_prob: float = 0.9
    clutter_density: float = 1e-6
    gate_probability: float = 0.99
    confirm_threshold: int = 3
    delete_threshold: int = 5
    min_hypothesis_prob: float = 1e-6
    new_track_weight: float = 0.1


class MHTResult(NamedTuple):
    """Result of MHT processing step.

    Attributes
    ----------
    confirmed_tracks : list of MHTTrack
        Tracks that are confirmed.
    tentative_tracks : list of MHTTrack
        Tracks that are tentative.
    all_tracks : list of MHTTrack
        All active tracks from best hypothesis.
    n_hypotheses : int
        Number of active hypotheses.
    best_hypothesis_prob : float
        Probability of the best hypothesis.
    """

    confirmed_tracks: List[MHTTrack]
    tentative_tracks: List[MHTTrack]
    all_tracks: List[MHTTrack]
    n_hypotheses: int
    best_hypothesis_prob: float


class MHTTracker:
    """
    Multiple Hypothesis Tracking (MHT) tracker.

    Maintains multiple hypotheses about measurement-to-track associations,
    with N-scan pruning for complexity control.

    Parameters
    ----------
    state_dim : int
        Dimension of state vector.
    meas_dim : int
        Dimension of measurement vector.
    F : callable or ndarray
        State transition matrix or function F(dt) -> ndarray.
    H : ndarray
        Measurement matrix.
    Q : callable or ndarray
        Process noise covariance or function Q(dt) -> ndarray.
    R : ndarray
        Measurement noise covariance.
    config : MHTConfig, optional
        Tracker configuration. Uses defaults if not provided.
    init_covariance : ndarray, optional
        Initial covariance for new tracks.

    Examples
    --------
    >>> import numpy as np
    >>> # Constant velocity model
    >>> F = lambda dt: np.array([[1, dt, 0, 0],
    ...                          [0, 1, 0, 0],
    ...                          [0, 0, 1, dt],
    ...                          [0, 0, 0, 1]])
    >>> H = np.array([[1, 0, 0, 0],
    ...               [0, 0, 1, 0]])
    >>> Q = lambda dt: 0.1 * np.eye(4)
    >>> R = np.eye(2) * 0.5
    >>> tracker = MHTTracker(4, 2, F, H, Q, R)
    >>> # Process measurements
    >>> measurements = [np.array([1, 2]), np.array([5, 6])]
    >>> result = tracker.process(measurements, dt=1.0)
    """

    def __init__(
        self,
        state_dim: int,
        meas_dim: int,
        F: Callable[[float], NDArray[np.floating]] | NDArray[np.floating],
        H: NDArray[np.floating],
        Q: Callable[[float], NDArray[np.floating]] | NDArray[np.floating],
        R: NDArray[np.floating],
        config: Optional[MHTConfig] = None,
        init_covariance: Optional[NDArray[np.floating]] = None,
    ):
        self.state_dim = state_dim
        self.meas_dim = meas_dim

        self._F = F if callable(F) else lambda dt: np.asarray(F, dtype=np.float64)
        self.H = np.asarray(H, dtype=np.float64)
        self._Q = Q if callable(Q) else lambda dt: np.asarray(Q, dtype=np.float64)
        self.R = np.asarray(R, dtype=np.float64)

        self.config = config or MHTConfig()

        if init_covariance is not None:
            self.init_covariance = np.asarray(init_covariance, dtype=np.float64)
        else:
            self.init_covariance = np.eye(state_dim) * 100.0

        # Compute gate threshold
        self.gate_threshold = chi2.ppf(self.config.gate_probability, df=meas_dim)

        # Initialize hypothesis tree
        self.hypothesis_tree = HypothesisTree(
            max_hypotheses=self.config.max_hypotheses,
            n_scan=self.config.n_scan,
            min_probability=self.config.min_hypothesis_prob,
        )
        self.hypothesis_tree.initialize()

        self._time = 0.0
        self._scan = 0

    def process(
        self,
        measurements: List[ArrayLike],
        dt: float,
    ) -> MHTResult:
        """
        Process measurements at new time step.

        Parameters
        ----------
        measurements : list of array_like
            List of measurement vectors.
        dt : float
            Time step since last update.

        Returns
        -------
        result : MHTResult
            Tracking result with confirmed and tentative tracks.
        """
        self._time += dt
        self._scan += 1

        # Convert measurements
        if len(measurements) == 0:
            Z = np.zeros((0, self.meas_dim))
        else:
            Z = np.array([np.asarray(m, dtype=np.float64) for m in measurements])

        n_meas = len(Z)

        # Get current tracks from all hypotheses
        all_track_ids = set()
        for hyp in self.hypothesis_tree.hypotheses:
            all_track_ids.update(hyp.track_ids)

        current_tracks = {
            tid: self.hypothesis_tree.tracks[tid]
            for tid in all_track_ids
            if tid in self.hypothesis_tree.tracks
        }

        # Predict all tracks
        F = self._F(dt)
        Q = self._Q(dt)
        predicted_tracks = self._predict_tracks(current_tracks, F, Q)

        # Compute gating and likelihoods
        gated, likelihood_matrix = self._compute_gating_and_likelihoods(
            predicted_tracks, Z
        )

        # Generate associations for each hypothesis
        track_id_list = list(predicted_tracks.keys())
        n_tracks = len(track_id_list)

        if n_tracks > 0:
            # Create gating matrix indexed by position
            gated_matrix = np.zeros((n_tracks, n_meas), dtype=bool)
            likelihood_mat = np.zeros((n_tracks, n_meas))
            for i, tid in enumerate(track_id_list):
                for j in range(n_meas):
                    if (tid, j) in gated:
                        gated_matrix[i, j] = True
                        likelihood_mat[i, j] = likelihood_matrix.get((tid, j), 0.0)

            # Generate joint associations
            associations = generate_joint_associations(gated_matrix, n_tracks, n_meas)
        else:
            associations = [{}]

        # Compute likelihood for each association
        assoc_likelihoods = []
        for assoc in associations:
            # Convert position-based to track_id-based
            track_assoc = {}
            for pos_idx, meas_idx in assoc.items():
                track_id = track_id_list[pos_idx] if n_tracks > 0 else -1
                track_assoc[track_id] = meas_idx

            # Compute likelihood
            lik = self._compute_association_likelihood(
                track_assoc, predicted_tracks, Z, likelihood_matrix
            )
            assoc_likelihoods.append(lik)

        # Normalize likelihoods
        total_lik = sum(assoc_likelihoods)
        if total_lik > 0:
            assoc_likelihoods = [lik / total_lik for lik in assoc_likelihoods]
        elif len(assoc_likelihoods) > 0:
            assoc_likelihoods = [1.0 / len(assoc_likelihoods)] * len(assoc_likelihoods)

        # Update tracks based on associations
        new_tracks_per_assoc: Dict[int, List[MHTTrack]] = {}
        updated_tracks: Dict[int, Dict[int, MHTTrack]] = (
            {}
        )  # assoc_idx -> track_id -> track

        for assoc_idx, assoc in enumerate(associations):
            updated_tracks[assoc_idx] = {}
            new_tracks_per_assoc[assoc_idx] = []

            # Update existing tracks
            for pos_idx, meas_idx in assoc.items():
                if n_tracks == 0:
                    continue
                track_id = track_id_list[pos_idx]
                pred_track = predicted_tracks[track_id]

                if meas_idx >= 0:
                    # Track with measurement
                    upd_track = self._update_track(pred_track, Z[meas_idx], meas_idx)
                else:
                    # Missed detection
                    upd_track = self._miss_track(pred_track)

                updated_tracks[assoc_idx][track_id] = upd_track

            # Handle unassigned measurements -> new tracks
            assigned_meas = set(
                meas_idx for meas_idx in assoc.values() if meas_idx >= 0
            )
            for j in range(n_meas):
                if j not in assigned_meas:
                    new_track = self._initiate_track(Z[j], j)
                    new_tracks_per_assoc[assoc_idx].append(new_track)

        # Store updated tracks
        for assoc_idx, track_dict in updated_tracks.items():
            for track_id, track in track_dict.items():
                self.hypothesis_tree.tracks[track_id] = track

        # Expand hypotheses
        self._expand_hypotheses(
            associations,
            assoc_likelihoods,
            new_tracks_per_assoc,
            track_id_list,
        )

        # Build result
        return self._build_result()

    def _predict_tracks(
        self,
        tracks: Dict[int, MHTTrack],
        F: NDArray,
        Q: NDArray,
    ) -> Dict[int, MHTTrack]:
        """Predict all tracks forward in time."""
        predicted = {}
        for tid, track in tracks.items():
            if track.status == MHTTrackStatus.DELETED:
                continue

            x_pred = F @ track.state
            P_pred = F @ track.covariance @ F.T + Q

            predicted[tid] = MHTTrack(
                id=track.id,
                state=x_pred,
                covariance=P_pred,
                score=track.score,
                status=track.status,
                history=track.history,
                parent_id=track.parent_id,
                scan_created=track.scan_created,
                n_hits=track.n_hits,
                n_misses=track.n_misses,
            )

        return predicted

    def _compute_gating_and_likelihoods(
        self,
        tracks: Dict[int, MHTTrack],
        Z: NDArray,
    ) -> tuple[Set[tuple], Dict[tuple, float]]:
        """Compute gating matrix and likelihood values."""
        gated = set()
        likelihood_matrix = {}

        for tid, track in tracks.items():
            z_pred = self.H @ track.state
            S = self.H @ track.covariance @ self.H.T + self.R

            for j in range(len(Z)):
                innovation = Z[j] - z_pred
                mahal_dist = mahalanobis_distance(innovation, S)

                if mahal_dist <= self.gate_threshold:
                    gated.add((tid, j))

                    # Compute likelihood
                    det_S = np.linalg.det(S)
                    if det_S > 0:
                        m = len(innovation)
                        mahal_sq = innovation @ np.linalg.solve(S, innovation)
                        likelihood = (
                            self.config.detection_prob
                            * np.exp(-0.5 * mahal_sq)
                            / np.sqrt((2 * np.pi) ** m * det_S)
                        )
                        likelihood_matrix[(tid, j)] = likelihood

        return gated, likelihood_matrix

    def _compute_association_likelihood(
        self,
        association: Dict[int, int],
        tracks: Dict[int, MHTTrack],
        Z: NDArray,
        likelihood_matrix: Dict[tuple, float],
    ) -> float:
        """Compute likelihood of a joint association."""
        likelihood = 1.0

        used_meas = set()
        for track_id, meas_idx in association.items():
            if meas_idx == -1:
                # Missed detection
                likelihood *= 1.0 - self.config.detection_prob
            else:
                # Detection
                if (track_id, meas_idx) in likelihood_matrix:
                    likelihood *= likelihood_matrix[(track_id, meas_idx)]
                else:
                    likelihood *= 1e-10  # Very small for ungated
                used_meas.add(meas_idx)

        # Clutter and new track terms for unassigned measurements
        n_unassigned = len(Z) - len(used_meas)
        likelihood *= (
            self.config.clutter_density + self.config.new_track_weight
        ) ** n_unassigned

        return likelihood

    def _update_track(
        self,
        track: MHTTrack,
        measurement: NDArray,
        meas_idx: int,
    ) -> MHTTrack:
        """Update a track with a measurement."""
        # Innovation
        z_pred = self.H @ track.state
        innovation = measurement - z_pred
        S = self.H @ track.covariance @ self.H.T + self.R

        # Kalman gain
        K = track.covariance @ self.H.T @ np.linalg.inv(S)

        # Update state and covariance
        x_upd = track.state + K @ innovation
        P_upd = (np.eye(self.state_dim) - K @ self.H) @ track.covariance

        # Update score (log-likelihood ratio)
        det_S = np.linalg.det(S)
        if det_S > 0:
            mahal_sq = innovation @ np.linalg.solve(S, innovation)
            score_update = 0.5 * (
                np.log(self.config.detection_prob)
                - np.log(self.config.clutter_density)
                - 0.5 * mahal_sq
                - 0.5 * np.log(det_S)
            )
        else:
            score_update = 0.0

        new_score = track.score + score_update

        # Update status
        n_hits = track.n_hits + 1
        n_misses = 0
        status = track.status
        if status == MHTTrackStatus.TENTATIVE:
            if n_hits >= self.config.confirm_threshold:
                status = MHTTrackStatus.CONFIRMED

        # Update history
        new_history = track.history + [meas_idx]

        return MHTTrack(
            id=track.id,
            state=x_upd,
            covariance=P_upd,
            score=new_score,
            status=status,
            history=new_history,
            parent_id=track.parent_id,
            scan_created=track.scan_created,
            n_hits=n_hits,
            n_misses=n_misses,
        )

    def _miss_track(self, track: MHTTrack) -> MHTTrack:
        """Handle missed detection for a track."""
        # Update score for missed detection
        score_update = np.log(1.0 - self.config.detection_prob)
        new_score = track.score + score_update

        # Update status
        n_misses = track.n_misses + 1
        status = track.status
        if n_misses >= self.config.delete_threshold:
            status = MHTTrackStatus.DELETED

        # Update history
        new_history = track.history + [-1]

        return MHTTrack(
            id=track.id,
            state=track.state,
            covariance=track.covariance,
            score=new_score,
            status=status,
            history=new_history,
            parent_id=track.parent_id,
            scan_created=track.scan_created,
            n_hits=track.n_hits,
            n_misses=n_misses,
        )

    def _initiate_track(
        self,
        measurement: NDArray,
        meas_idx: int,
    ) -> MHTTrack:
        """Initiate a new track from a measurement."""
        # Initialize state from measurement
        H_pinv = np.linalg.pinv(self.H)
        state = H_pinv @ measurement

        return MHTTrack(
            id=-1,  # Will be assigned by hypothesis tree
            state=state,
            covariance=self.init_covariance.copy(),
            score=np.log(self.config.new_track_weight),
            status=MHTTrackStatus.TENTATIVE,
            history=[meas_idx],
            parent_id=-1,
            scan_created=self._scan,
            n_hits=1,
            n_misses=0,
        )

    def _expand_hypotheses(
        self,
        associations: List[Dict[int, int]],
        likelihoods: List[float],
        new_tracks: Dict[int, List[MHTTrack]],
        track_id_list: List[int],
    ) -> None:
        """Expand hypotheses with new associations."""
        new_hypotheses = []

        for hyp in self.hypothesis_tree.hypotheses:
            for assoc_idx, (assoc, likelihood) in enumerate(
                zip(associations, likelihoods)
            ):
                # Compute new hypothesis probability
                new_prob = hyp.probability * likelihood

                if new_prob < self.config.min_hypothesis_prob:
                    continue

                # Determine track IDs for new hypothesis
                new_track_ids = []

                # Keep surviving tracks from original hypothesis
                for track_id in hyp.track_ids:
                    if track_id in self.hypothesis_tree.tracks:
                        track = self.hypothesis_tree.tracks[track_id]
                        if track.status != MHTTrackStatus.DELETED:
                            new_track_ids.append(track_id)

                # Add new tracks from this association
                if assoc_idx in new_tracks:
                    for new_track in new_tracks[assoc_idx]:
                        tid = self.hypothesis_tree.add_track(new_track)
                        new_track_ids.append(tid)

                # Create new hypothesis
                new_hyp = Hypothesis(
                    id=self.hypothesis_tree._get_next_hypothesis_id(),
                    probability=new_prob,
                    track_ids=new_track_ids,
                    scan_created=self._scan,
                    parent_id=hyp.id,
                )
                new_hypotheses.append(new_hyp)

        # Update hypotheses
        if new_hypotheses:
            self.hypothesis_tree.hypotheses = new_hypotheses
        else:
            # Keep at least one hypothesis
            if self.hypothesis_tree.hypotheses:
                best = max(self.hypothesis_tree.hypotheses, key=lambda h: h.probability)
                self.hypothesis_tree.hypotheses = [best]

        # Prune
        self.hypothesis_tree.prune()

    def _build_result(self) -> MHTResult:
        """Build result from current state."""
        best_tracks = self.hypothesis_tree.get_best_tracks()
        confirmed = [t for t in best_tracks if t.status == MHTTrackStatus.CONFIRMED]
        tentative = [t for t in best_tracks if t.status == MHTTrackStatus.TENTATIVE]

        best_hyp = self.hypothesis_tree.get_best_hypothesis()
        best_prob = best_hyp.probability if best_hyp else 0.0

        return MHTResult(
            confirmed_tracks=confirmed,
            tentative_tracks=tentative,
            all_tracks=best_tracks,
            n_hypotheses=len(self.hypothesis_tree.hypotheses),
            best_hypothesis_prob=best_prob,
        )

    @property
    def tracks(self) -> List[MHTTrack]:
        """Get all tracks from best hypothesis."""
        return self.hypothesis_tree.get_best_tracks()

    @property
    def confirmed_tracks(self) -> List[MHTTrack]:
        """Get confirmed tracks from best hypothesis."""
        return self.hypothesis_tree.get_confirmed_tracks()

    @property
    def n_hypotheses(self) -> int:
        """Number of active hypotheses."""
        return len(self.hypothesis_tree.hypotheses)


__all__ = [
    "MHTConfig",
    "MHTResult",
    "MHTTracker",
]



"""
Tests for INS/GNSS Integration.

Tests cover:
- GNSS measurement models
- Loosely-coupled integration
- Tightly-coupled integration
- DOP computation
- Fault detection
"""

import numpy as np
import pytest

from pytcl.navigation.geodesy import geodetic_to_ecef
from pytcl.navigation.ins import IMUData, initialize_ins_state, normal_gravity
from pytcl.navigation.ins_gnss import (
    GPS_L1_FREQ,
    GPS_L1_WAVELENGTH,
    SPEED_OF_LIGHT,
    GNSSMeasurement,
    INSGNSSState,
    SatelliteInfo,
    compute_dop,
    compute_line_of_sight,
    gnss_outage_detection,
    initialize_ins_gnss,
    loose_coupled_predict,
    loose_coupled_update,
    loose_coupled_update_position,
    loose_coupled_update_velocity,
    position_measurement_matrix,
    position_velocity_measurement_matrix,
    pseudorange_measurement_matrix,
    satellite_elevation_azimuth,
    tight_coupled_pseudorange_innovation,
    tight_coupled_update,
    velocity_measurement_matrix,
)


class TestConstants:
    """Tests for GNSS constants."""

    def test_speed_of_light(self):
        """Test speed of light constant."""
        assert SPEED_OF_LIGHT == pytest.approx(299792458.0, rel=1e-10)

    def test_gps_l1_frequency(self):
        """Test GPS L1 frequency."""
        assert GPS_L1_FREQ == pytest.approx(1575.42e6, rel=1e-6)

    def test_gps_l1_wavelength(self):
        """Test GPS L1 wavelength consistency."""
        expected = SPEED_OF_LIGHT / GPS_L1_FREQ
        assert GPS_L1_WAVELENGTH == pytest.approx(expected, rel=1e-10)


class TestMeasurementMatrices:
    """Tests for measurement matrix construction."""

    def test_position_measurement_matrix_shape(self):
        """Test position measurement matrix shape."""
        H = position_measurement_matrix()
        assert H.shape == (3, 15)

    def test_position_measurement_matrix_values(self):
        """Test position measurement matrix extracts position errors."""
        H = position_measurement_matrix()

        # First 3 elements of error state are position errors
        error_state = np.zeros(15)
        error_state[0] = 1e-6  # lat error
        error_state[1] = 2e-6  # lon error
        error_state[2] = 10.0  # alt error

        result = H @ error_state
        np.testing.assert_allclose(result, [1e-6, 2e-6, 10.0])

    def test_velocity_measurement_matrix_shape(self):
        """Test velocity measurement matrix shape."""
        H = velocity_measurement_matrix()
        assert H.shape == (3, 15)

    def test_velocity_measurement_matrix_values(self):
        """Test velocity measurement matrix extracts velocity errors."""
        H = velocity_measurement_matrix()

        error_state = np.zeros(15)
        error_state[3] = 1.0  # vN error
        error_state[4] = 2.0  # vE error
        error_state[5] = 0.5  # vD error

        result = H @ error_state
        np.testing.assert_allclose(result, [1.0, 2.0, 0.5])

    def test_position_velocity_measurement_matrix_shape(self):
        """Test combined measurement matrix shape."""
        H = position_velocity_measurement_matrix()
        assert H.shape == (6, 15)

    def test_position_velocity_measurement_matrix_values(self):
        """Test combined matrix extracts both position and velocity."""
        H = position_velocity_measurement_matrix()

        error_state = np.zeros(15)
        error_state[0] = 1e-6
        error_state[3] = 1.0

        result = H @ error_state
        assert result[0] == pytest.approx(1e-6)
        assert result[3] == pytest.approx(1.0)


class TestLineOfSight:
    """Tests for line-of-sight computation."""

    def test_line_of_sight_vertical(self):
        """Test LOS for satellite directly above."""
        user = np.array([6378137.0, 0.0, 0.0])  # On equator
        sat = np.array([6378137.0 + 20200000.0, 0.0, 0.0])  # Directly above

        los, range_val = compute_line_of_sight(user, sat)

        # LOS should point radially outward
        np.testing.assert_allclose(los, [1.0, 0.0, 0.0], atol=1e-10)
        assert range_val == pytest.approx(20200000.0, rel=1e-10)

    def test_line_of_sight_unit_vector(self):
        """Test LOS is a unit vector."""
        user = np.array([6378137.0, 0.0, 0.0])
        sat = np.array([6378137.0 + 20000000.0, 5000000.0, 3000000.0])

        los, _ = compute_line_of_sight(user, sat)

        assert np.linalg.norm(los) == pytest.approx(1.0, rel=1e-10)


class TestPseudorangeGeometry:
    """Tests for pseudorange measurement geometry."""

    def test_pseudorange_matrix_shape(self):
        """Test pseudorange measurement matrix shape."""
        user = np.array([6378137.0, 0.0, 0.0])
        satellites = [
            SatelliteInfo(
                prn=i,
                position=np.array([6378137.0 + 20200000.0, i * 1e6, 0.0]),
                velocity=np.zeros(3),
                pseudorange=22000000.0,
            )
            for i in range(4)
        ]

        H = pseudorange_measurement_matrix(user, satellites, include_clock=True)
        assert H.shape == (4, 4)

        H_no_clock = pseudorange_measurement_matrix(
            user, satellites, include_clock=False
        )
        assert H_no_clock.shape == (4, 3)

    def test_pseudorange_matrix_clock_column(self):
        """Test clock bias column is all ones."""
        user = np.array([6378137.0, 0.0, 0.0])
        satellites = [
            SatelliteInfo(
                prn=i,
                position=np.array([6378137.0 + 20200000.0, i * 1e6, 0.0]),
                velocity=np.zeros(3),
                pseudorange=22000000.0,
            )
            for i in range(4)
        ]

        H = pseudorange_measurement_matrix(user, satellites, include_clock=True)

        # Clock column should be all ones
        np.testing.assert_allclose(H[:, 3], np.ones(4), atol=1e-10)


class TestDOP:
    """Tests for Dilution of Precision computation."""

    def test_dop_good_geometry(self):
        """Test DOP with good satellite geometry."""
        # Create geometry matrix with good coverage (diverse LOS vectors)
        # Satellites at different azimuths and elevations
        H = np.array(
            [
                [0.5, 0.5, 0.7071, 1.0],  # NE, high
                [-0.5, 0.5, 0.7071, 1.0],  # NW, high
                [0.5, -0.5, 0.7071, 1.0],  # SE, high
                [-0.5, -0.5, 0.7071, 1.0],  # SW, high
                [0.0, 0.0, 1.0, 1.0],  # Directly overhead (adds vertical diversity)
            ]
        )

        gdop, pdop, hdop, vdop = compute_dop(H)

        # All DOP values should be finite and reasonable
        assert np.isfinite(gdop)
        assert np.isfinite(pdop)
        assert np.isfinite(hdop)
        assert np.isfinite(vdop)

        # PDOP should be less than GDOP (GDOP includes clock)
        assert pdop < gdop

    def test_dop_poor_geometry(self):
        """Test DOP with poor (coplanar) geometry."""
        # Coplanar satellites - poor vertical geometry
        H = np.array(
            [
                [1.0, 0.0, 0.0, 1.0],
                [0.0, 1.0, 0.0, 1.0],
                [-1.0, 0.0, 0.0, 1.0],
                [0.0, -1.0, 0.0, 1.0],
            ]
        )

        gdop, pdop, hdop, vdop = compute_dop(H)

        # VDOP should be very large (poor vertical geometry)
        assert vdop > 10 or np.isinf(vdop)


class TestSatelliteElevationAzimuth:
    """Tests for satellite elevation/azimuth computation."""

    def test_satellite_directly_above(self):
        """Test elevation for satellite directly above."""
        user_lla = np.array([0.0, 0.0, 0.0])  # Equator, prime meridian

        # Satellite directly above on x-axis
        sat_ecef = np.array([6378137.0 + 20200000.0, 0.0, 0.0])

        elev, az = satellite_elevation_azimuth(user_lla, sat_ecef)

        # Should be at 90 degrees elevation
        assert elev == pytest.approx(np.pi / 2, abs=0.01)

    def test_satellite_on_horizon(self):
        """Test elevation for satellite on horizon."""
        user_lla = np.array([0.0, 0.0, 0.0])  # Equator, prime meridian

        # Satellite to the east, same altitude as user (approximately on horizon)
        # This is a simplified test - actual horizon geometry is more complex
        sat_ecef = np.array([6378137.0, 20200000.0, 0.0])

        elev, az = satellite_elevation_azimuth(user_lla, sat_ecef)

        # Should be close to 90 degrees azimuth (east)
        assert az == pytest.approx(np.pi / 2, abs=0.1)

    def test_azimuth_range(self):
        """Test azimuth is in valid range [0, 2pi)."""
        user_lla = np.array([np.radians(45), np.radians(-75), 100.0])

        for angle in np.linspace(0, 2 * np.pi, 8):
            sat_ecef = np.array(
                [
                    6378137.0 + 20200000.0 * np.cos(angle),
                    20200000.0 * np.sin(angle),
                    10000000.0,
                ]
            )
            _, az = satellite_elevation_azimuth(user_lla, sat_ecef)
            assert 0 <= az < 2 * np.pi


class TestINSGNSSInitialization:
    """Tests for INS/GNSS state initialization."""

    def test_initialize_ins_gnss(self):
        """Test basic initialization."""
        ins_state = initialize_ins_state(lat=np.radians(45), lon=0, alt=100)

        state = initialize_ins_gnss(ins_state)

        assert isinstance(state, INSGNSSState)
        assert state.ins_state == ins_state
        np.testing.assert_allclose(state.error_state, np.zeros(15))
        assert state.error_cov.shape == (15, 15)
        assert state.clock_bias == 0.0
        assert state.clock_drift == 0.0

    def test_initialize_covariance_diagonal(self):
        """Test initial covariance is diagonal."""
        ins_state = initialize_ins_state(lat=0, lon=0, alt=0)

        state = initialize_ins_gnss(
            ins_state,
            position_std=5.0,
            velocity_std=0.5,
        )

        # Check diagonal elements match expected variances
        assert state.error_cov[0, 0] == pytest.approx(25.0)  # position variance
        assert state.error_cov[3, 3] == pytest.approx(0.25)  # velocity variance

        # Off-diagonal should be zero
        assert state.error_cov[0, 3] == 0.0


class TestLooseCoupledIntegration:
    """Tests for loosely-coupled INS/GNSS integration."""

    def test_loose_coupled_predict(self):
        """Test prediction step."""
        ins_state = initialize_ins_state(lat=np.radians(45), lon=0, alt=1000)
        state = initialize_ins_gnss(ins_state)

        g = normal_gravity(np.radians(45), 1000)
        imu = IMUData(
            accel=np.array([0.0, 0.0, -g]),
            gyro=np.array([0.0, 0.0, 0.0]),
            dt=0.01,
        )

        new_state = loose_coupled_predict(state, imu)

        # State should be updated
        assert new_state.ins_state.time == state.ins_state.time + 0.01

        # Covariance should grow during prediction
        assert np.trace(new_state.error_cov) >= np.trace(state.error_cov)

    def test_loose_coupled_update_position(self):
        """Test position update reduces position uncertainty."""
        ins_state = initialize_ins_state(lat=np.radians(45), lon=0, alt=1000)
        state = initialize_ins_gnss(ins_state, position_std=100.0)

        # GNSS measurement at true position
        gnss = GNSSMeasurement(
            position=ins_state.position,
            velocity=None,
            position_cov=np.diag([1.0, 1.0, 2.0]),
            velocity_cov=None,
            time=0.0,
            valid=True,
        )

        result = loose_coupled_update_position(state, gnss)

        # Position covariance should decrease
        pos_var_before = np.trace(state.error_cov[0:3, 0:3])
        pos_var_after = np.trace(result.state.error_cov[0:3, 0:3])
        assert pos_var_after < pos_var_before

    def test_loose_coupled_update_velocity(self):
        """Test velocity update reduces velocity uncertainty."""
        ins_state = initialize_ins_state(lat=np.radians(45), lon=0, alt=1000, vN=10)
        state = initialize_ins_gnss(ins_state, velocity_std=10.0)

        gnss = GNSSMeasurement(
            position=None,
            velocity=ins_state.velocity,
            position_cov=None,
            velocity_cov=np.diag([0.1, 0.1, 0.1]),
            time=0.0,
            valid=True,
        )

        result = loose_coupled_update_velocity(state, gnss)

        vel_var_before = np.trace(state.error_cov[3:6, 3:6])
        vel_var_after = np.trace(result.state.error_cov[3:6, 3:6])
        assert vel_var_after < vel_var_before

    def test_loose_coupled_update_combined(self):
        """Test combined position+velocity update."""
        ins_state = initialize_ins_state(lat=np.radians(45), lon=0, alt=1000, vN=10)
        state = initialize_ins_gnss(ins_state)

        gnss = GNSSMeasurement(
            position=ins_state.position + np.array([1e-6, 0, 5]),  # Small offset
            velocity=ins_state.velocity + np.array([0.5, 0, 0]),
            position_cov=np.diag([1.0, 1.0, 2.0]),
            velocity_cov=np.diag([0.1, 0.1, 0.1]),
            time=0.0,
            valid=True,
        )

        result = loose_coupled_update(state, gnss)

        # Innovation should reflect the measurement offset
        assert result.innovation.shape == (6,)
        assert result.innovation_cov.shape == (6, 6)

    def test_loose_coupled_invalid_measurement(self):
        """Test handling of invalid GNSS measurement."""
        ins_state = initialize_ins_state(lat=np.radians(45), lon=0, alt=1000)
        state = initialize_ins_gnss(ins_state)

        gnss = GNSSMeasurement(
            position=None,
            velocity=None,
            position_cov=None,
            velocity_cov=None,
            time=0.0,
            valid=False,
        )

        result = loose_coupled_update(state, gnss)

        # State should be unchanged
        np.testing.assert_allclose(
            result.state.ins_state.position, state.ins_state.position
        )


class TestTightCoupledIntegration:
    """Tests for tightly-coupled INS/GNSS integration."""

    def _create_test_satellites(self, user_ecef):
        """Create test satellite constellation with good geometry."""
        # Create satellites at various positions with diverse geometry
        # Including vertical diversity for good VDOP
        satellites = []
        sat_positions = [
            (20e6, 5e6, 10e6),  # NE, high
            (20e6, -5e6, 10e6),  # SE, high
            (-10e6, 10e6, 15e6),  # NW, higher
            (-10e6, -10e6, 15e6),  # SW, higher
            (5e6, 0, 25e6),  # Near zenith
        ]
        for i, (dx, dy, dz) in enumerate(sat_positions):
            sat_pos = user_ecef + np.array([dx, dy, dz])
            geo_range = np.linalg.norm(sat_pos - user_ecef)

            satellites.append(
                SatelliteInfo(
                    prn=i + 1,
                    position=sat_pos,
                    velocity=np.zeros(3),
                    pseudorange=geo_range + 100.0,  # Include clock bias
                )
            )
        return satellites

    def test_tight_coupled_innovation(self):
        """Test pseudorange innovation computation."""
        ins_state = initialize_ins_state(lat=np.radians(45), lon=0, alt=1000)
        state = initialize_ins_gnss(ins_state)

        # Get user ECEF position
        lat, lon, alt = ins_state.position
        user_x, user_y, user_z = geodetic_to_ecef(lat, lon, alt)
        user_ecef = np.array([user_x, user_y, user_z])

        satellites = self._create_test_satellites(user_ecef)

        innovations, predicted = tight_coupled_pseudorange_innovation(state, satellites)

        assert innovations.shape == (5,)
        assert predicted.shape == (5,)

        # Innovations should be approximately the clock bias (100m in our test)
        for innov in innovations:
            assert innov == pytest.approx(100.0, abs=1.0)

    def test_tight_coupled_update(self):
        """Test tightly-coupled update with pseudoranges."""
        ins_state = initialize_ins_state(lat=np.radians(45), lon=0, alt=1000)
        state = initialize_ins_gnss(ins_state)

        lat, lon, alt = ins_state.position
        user_x, user_y, user_z = geodetic_to_ecef(lat, lon, alt)
        user_ecef = np.array([user_x, user_y, user_z])

        satellites = self._create_test_satellites(user_ecef)

        result = tight_coupled_update(state, satellites, pseudorange_std=3.0)

        # Should have valid DOP values
        gdop, pdop, hdop, vdop = result.dop
        assert np.isfinite(gdop)
        assert np.isfinite(pdop)

        # Clock bias should be estimated
        assert result.state.clock_bias != 0.0

    def test_tight_coupled_insufficient_satellites(self):
        """Test handling of insufficient satellites."""
        ins_state = initialize_ins_state(lat=np.radians(45), lon=0, alt=1000)
        state = initialize_ins_gnss(ins_state)

        # Only 3 satellites (need at least 4)
        satellites = [
            SatelliteInfo(
                prn=i,
                position=np.array([20e6, i * 1e6, 0]),
                velocity=np.zeros(3),
                pseudorange=22e6,
            )
            for i in range(3)
        ]

        result = tight_coupled_update(state, satellites)

        # Should return infinite DOP
        assert np.isinf(result.dop[0])


class TestFaultDetection:
    """Tests for GNSS fault detection."""

    def test_no_fault_detected(self):
        """Test no fault for consistent measurements."""
        innovations = np.array([0.5, 0.3, -0.2])
        innovation_cov = np.eye(3) * 1.0

        fault = gnss_outage_detection(innovations, innovation_cov, threshold=10.0)
        assert not fault

    def test_fault_detected_large_innovation(self):
        """Test fault detected for large innovations."""
        innovations = np.array([10.0, 10.0, 10.0])  # Large innovations
        innovation_cov = np.eye(3) * 0.1  # Small expected variance

        fault = gnss_outage_detection(innovations, innovation_cov, threshold=5.991)
        assert fault

    def test_fault_singular_covariance(self):
        """Test fault returned for singular covariance."""
        innovations = np.array([0.1, 0.1])
        innovation_cov = np.zeros((2, 2))  # Singular

        fault = gnss_outage_detection(innovations, innovation_cov)
        assert fault


class TestIntegration:
    """Integration tests for INS/GNSS."""

    def test_predict_update_cycle(self):
        """Test complete predict-update cycle."""
        ins_state = initialize_ins_state(
            lat=np.radians(45), lon=np.radians(-75), alt=100, vN=10
        )
        state = initialize_ins_gnss(ins_state)

        g = normal_gravity(np.radians(45), 100)

        # Run several predict-update cycles
        for i in range(10):
            # IMU prediction at 100 Hz
            for _ in range(10):
                imu = IMUData(
                    accel=np.array([0.0, 0.0, -g]),
                    gyro=np.array([0.0, 0.0, 0.0]),
                    dt=0.01,
                )
                state = loose_coupled_predict(state, imu)

            # GNSS update at 10 Hz
            gnss = GNSSMeasurement(
                position=state.ins_state.position,  # Perfect measurement
                velocity=state.ins_state.velocity,
                position_cov=np.diag([1.0, 1.0, 2.0]),
                velocity_cov=np.diag([0.1, 0.1, 0.1]),
                time=float(i) * 0.1,
                valid=True,
            )
            result = loose_coupled_update(state, gnss)
            state = result.state

        # Final state should be stable
        assert np.isfinite(state.ins_state.position).all()
        assert np.isfinite(state.error_cov).all()

        # Covariance should be bounded
        assert np.trace(state.error_cov) < 1e6

    def test_gnss_outage_recovery(self):
        """Test system behavior during GNSS outage."""
        ins_state = initialize_ins_state(lat=np.radians(45), lon=0, alt=100)
        state = initialize_ins_gnss(ins_state)

        initial_cov = np.trace(state.error_cov)

        g = normal_gravity(np.radians(45), 100)

        # Simulate GNSS outage (predictions only)
        for _ in range(100):
            imu = IMUData(
                accel=np.array([0.0, 0.0, -g]),
                gyro=np.array([0.0, 0.0, 0.0]),
                dt=0.01,
            )
            state = loose_coupled_predict(state, imu)

        # Covariance should grow during outage
        assert np.trace(state.error_cov) > initial_cov

        # GNSS recovery
        gnss = GNSSMeasurement(
            position=state.ins_state.position,
            velocity=state.ins_state.velocity,
            position_cov=np.diag([1.0, 1.0, 2.0]),
            velocity_cov=np.diag([0.1, 0.1, 0.1]),
            time=1.0,
            valid=True,
        )
        result = loose_coupled_update(state, gnss)

        # Covariance should decrease after GNSS update
        assert np.trace(result.state.error_cov) < np.trace(state.error_cov)



"""Tests for rhumb line navigation functions."""

import numpy as np
import pytest

from pytcl.navigation.rhumb import (
    RhumbDirectResult,
    RhumbResult,
    compare_great_circle_rhumb,
    direct_rhumb,
    direct_rhumb_spherical,
    indirect_rhumb,
    indirect_rhumb_spherical,
    rhumb_bearing,
    rhumb_distance_ellipsoidal,
    rhumb_distance_spherical,
    rhumb_intersect,
    rhumb_midpoint,
    rhumb_waypoints,
)


class TestRhumbDistanceSpherical:
    """Tests for rhumb_distance_spherical."""

    def test_zero_distance(self):
        """Same point should have zero distance."""
        lat, lon = np.radians(40), np.radians(-74)
        dist = rhumb_distance_spherical(lat, lon, lat, lon)
        assert dist == pytest.approx(0.0, abs=1e-6)

    def test_due_north(self):
        """Due north rhumb should equal great circle distance."""
        lat1, lon1 = np.radians(40), np.radians(-74)
        lat2, lon2 = np.radians(50), np.radians(-74)

        rhumb_dist = rhumb_distance_spherical(lat1, lon1, lat2, lon2)

        # For same longitude, rhumb = great circle
        from pytcl.navigation.great_circle import great_circle_distance

        gc_dist = great_circle_distance(lat1, lon1, lat2, lon2)

        assert rhumb_dist == pytest.approx(gc_dist, rel=0.001)

    def test_due_east_at_equator(self):
        """Due east at equator rhumb equals great circle."""
        lat1, lon1 = np.radians(0), np.radians(0)
        lat2, lon2 = np.radians(0), np.radians(10)

        rhumb_dist = rhumb_distance_spherical(lat1, lon1, lat2, lon2)

        from pytcl.navigation.great_circle import great_circle_distance

        gc_dist = great_circle_distance(lat1, lon1, lat2, lon2)

        assert rhumb_dist == pytest.approx(gc_dist, rel=0.001)

    def test_rhumb_longer_than_great_circle(self):
        """Rhumb line should be longer than great circle for diagonal paths."""
        # New York to London
        lat1, lon1 = np.radians(40.7128), np.radians(-74.0060)
        lat2, lon2 = np.radians(51.5074), np.radians(-0.1278)

        rhumb_dist = rhumb_distance_spherical(lat1, lon1, lat2, lon2)

        from pytcl.navigation.great_circle import great_circle_distance

        gc_dist = great_circle_distance(lat1, lon1, lat2, lon2)

        assert rhumb_dist > gc_dist

    def test_wraparound_longitude(self):
        """Should handle longitude wraparound."""
        lat1, lon1 = np.radians(0), np.radians(170)
        lat2, lon2 = np.radians(0), np.radians(-170)

        dist = rhumb_distance_spherical(lat1, lon1, lat2, lon2)

        # Should cross date line (20 degrees), not go around (340 degrees)
        expected_approx = 6371000 * np.radians(20)  # ~2200 km
        assert dist == pytest.approx(expected_approx, rel=0.01)


class TestRhumbBearing:
    """Tests for rhumb_bearing."""

    def test_due_north(self):
        """Due north should have bearing 0."""
        lat1, lon1 = np.radians(40), np.radians(-74)
        lat2, lon2 = np.radians(50), np.radians(-74)

        bearing = rhumb_bearing(lat1, lon1, lat2, lon2)
        assert bearing == pytest.approx(0.0, abs=1e-6)

    def test_due_east(self):
        """Due east should have bearing π/2."""
        lat1, lon1 = np.radians(40), np.radians(-74)
        lat2, lon2 = np.radians(40), np.radians(-64)

        bearing = rhumb_bearing(lat1, lon1, lat2, lon2)
        assert bearing == pytest.approx(np.pi / 2, abs=1e-6)

    def test_due_south(self):
        """Due south should have bearing π."""
        lat1, lon1 = np.radians(40), np.radians(-74)
        lat2, lon2 = np.radians(30), np.radians(-74)

        bearing = rhumb_bearing(lat1, lon1, lat2, lon2)
        assert bearing == pytest.approx(np.pi, abs=1e-6)

    def test_due_west(self):
        """Due west should have bearing 3π/2."""
        lat1, lon1 = np.radians(40), np.radians(-74)
        lat2, lon2 = np.radians(40), np.radians(-84)

        bearing = rhumb_bearing(lat1, lon1, lat2, lon2)
        assert bearing == pytest.approx(3 * np.pi / 2, abs=1e-6)

    def test_normalized_range(self):
        """Bearing should be in [0, 2π)."""
        lat1, lon1 = np.radians(40), np.radians(-74)
        lat2, lon2 = np.radians(51), np.radians(-0.1)

        bearing = rhumb_bearing(lat1, lon1, lat2, lon2)
        assert 0 <= bearing < 2 * np.pi


class TestIndirectRhumbSpherical:
    """Tests for indirect_rhumb_spherical."""

    def test_returns_named_tuple(self):
        """Should return RhumbResult."""
        lat1, lon1 = np.radians(40), np.radians(-74)
        lat2, lon2 = np.radians(51), np.radians(-0.1)

        result = indirect_rhumb_spherical(lat1, lon1, lat2, lon2)

        assert isinstance(result, RhumbResult)
        assert hasattr(result, "distance")
        assert hasattr(result, "bearing")


class TestDirectRhumbSpherical:
    """Tests for direct_rhumb_spherical."""

    def test_zero_distance(self):
        """Zero distance should return starting point."""
        lat, lon = np.radians(40), np.radians(-74)
        bearing = np.radians(45)

        dest = direct_rhumb_spherical(lat, lon, bearing, 0.0)

        assert dest.lat == pytest.approx(lat, abs=1e-10)
        assert dest.lon == pytest.approx(lon, abs=1e-10)

    def test_due_north(self):
        """Moving north should increase latitude only."""
        lat, lon = np.radians(40), np.radians(-74)
        bearing = 0.0

        dest = direct_rhumb_spherical(lat, lon, bearing, 1000000)

        assert dest.lat > lat
        assert dest.lon == pytest.approx(lon, abs=1e-6)

    def test_due_east(self):
        """Moving east should change longitude only."""
        lat, lon = np.radians(40), np.radians(-74)
        bearing = np.pi / 2

        dest = direct_rhumb_spherical(lat, lon, bearing, 1000000)

        assert dest.lat == pytest.approx(lat, abs=1e-6)
        assert dest.lon > lon

    def test_roundtrip_consistency(self):
        """Direct then indirect should give same values."""
        lat1, lon1 = np.radians(40), np.radians(-74)
        bearing = np.radians(45)
        dist = 1000000

        dest = direct_rhumb_spherical(lat1, lon1, bearing, dist)
        result = indirect_rhumb_spherical(lat1, lon1, dest.lat, dest.lon)

        assert result.distance == pytest.approx(dist, rel=1e-4)
        assert result.bearing == pytest.approx(bearing, abs=1e-4)


class TestRhumbDistanceEllipsoidal:
    """Tests for rhumb_distance_ellipsoidal."""

    def test_zero_distance(self):
        """Same point should have zero distance."""
        lat, lon = np.radians(40), np.radians(-74)
        dist = rhumb_distance_ellipsoidal(lat, lon, lat, lon)
        assert dist == pytest.approx(0.0, abs=1)

    def test_similar_to_spherical(self):
        """Ellipsoidal should be close to spherical for short distances."""
        lat1, lon1 = np.radians(40), np.radians(-74)
        lat2, lon2 = np.radians(41), np.radians(-73)

        sph_dist = rhumb_distance_spherical(lat1, lon1, lat2, lon2)
        ell_dist = rhumb_distance_ellipsoidal(lat1, lon1, lat2, lon2)

        # Should be within a few percent
        assert ell_dist == pytest.approx(sph_dist, rel=0.02)


class TestIndirectRhumb:
    """Tests for indirect_rhumb (ellipsoidal)."""

    def test_returns_named_tuple(self):
        """Should return RhumbResult."""
        lat1, lon1 = np.radians(40), np.radians(-74)
        lat2, lon2 = np.radians(51), np.radians(-0.1)

        result = indirect_rhumb(lat1, lon1, lat2, lon2)

        assert isinstance(result, RhumbResult)


class TestDirectRhumb:
    """Tests for direct_rhumb (ellipsoidal)."""

    def test_returns_named_tuple(self):
        """Should return RhumbDirectResult."""
        lat, lon = np.radians(40), np.radians(-74)
        bearing = np.radians(45)

        dest = direct_rhumb(lat, lon, bearing, 1000000)

        assert isinstance(dest, RhumbDirectResult)

    def test_roundtrip_consistency(self):
        """Direct then indirect should give same values."""
        lat1, lon1 = np.radians(40), np.radians(-74)
        bearing = np.radians(45)
        dist = 1000000

        dest = direct_rhumb(lat1, lon1, bearing, dist)
        result = indirect_rhumb(lat1, lon1, dest.lat, dest.lon)

        assert result.distance == pytest.approx(dist, rel=0.01)
        assert result.bearing == pytest.approx(bearing, abs=0.01)


class TestRhumbIntersect:
    """Tests for rhumb_intersect."""

    def test_perpendicular_rhumbs(self):
        """Perpendicular rhumb lines should intersect."""
        lat1, lon1 = np.radians(40), np.radians(-74)
        bearing1 = np.radians(45)  # Northeast

        lat2, lon2 = np.radians(40), np.radians(-70)
        bearing2 = np.radians(315)  # Northwest

        result = rhumb_intersect(lat1, lon1, bearing1, lat2, lon2, bearing2)

        # Should intersect somewhere
        assert result.valid

    def test_crossing_rhumbs(self):
        """Two crossing rhumb lines."""
        lat1, lon1 = np.radians(0), np.radians(0)
        bearing1 = np.radians(45)  # Northeast

        lat2, lon2 = np.radians(10), np.radians(0)
        bearing2 = np.radians(135)  # Southeast

        result = rhumb_intersect(lat1, lon1, bearing1, lat2, lon2, bearing2)

        assert result.valid

    def test_parallel_rhumbs_no_intersect(self):
        """Parallel rhumb lines should not intersect."""
        lat1, lon1 = np.radians(0), np.radians(0)
        bearing1 = np.radians(45)

        lat2, lon2 = np.radians(0), np.radians(10)
        bearing2 = np.radians(45)  # Same bearing

        result = rhumb_intersect(lat1, lon1, bearing1, lat2, lon2, bearing2)

        assert not result.valid


class TestRhumbMidpoint:
    """Tests for rhumb_midpoint."""

    def test_midpoint(self):
        """Midpoint should be at half distance."""
        lat1, lon1 = np.radians(40), np.radians(-74)
        lat2, lon2 = np.radians(50), np.radians(-64)

        mid = rhumb_midpoint(lat1, lon1, lat2, lon2)

        # Distance from start to mid should equal mid to end
        d1 = rhumb_distance_spherical(lat1, lon1, mid.lat, mid.lon)
        d2 = rhumb_distance_spherical(mid.lat, mid.lon, lat2, lon2)

        assert d1 == pytest.approx(d2, rel=0.01)


class TestRhumbWaypoints:
    """Tests for rhumb_waypoints."""

    def test_n_points(self):
        """Should return correct number of points."""
        lat1, lon1 = np.radians(40), np.radians(-74)
        lat2, lon2 = np.radians(50), np.radians(-64)

        lats, lons = rhumb_waypoints(lat1, lon1, lat2, lon2, 10)

        assert len(lats) == 10
        assert len(lons) == 10

    def test_endpoints(self):
        """First and last points should match endpoints."""
        lat1, lon1 = np.radians(40), np.radians(-74)
        lat2, lon2 = np.radians(50), np.radians(-64)

        lats, lons = rhumb_waypoints(lat1, lon1, lat2, lon2, 5)

        assert lats[0] == pytest.approx(lat1, abs=1e-6)
        assert lons[0] == pytest.approx(lon1, abs=1e-6)
        assert lats[-1] == pytest.approx(lat2, abs=1e-4)
        assert lons[-1] == pytest.approx(lon2, abs=1e-4)


class TestCompareGreatCircleRhumb:
    """Tests for compare_great_circle_rhumb."""

    def test_returns_three_values(self):
        """Should return gc_distance, rhumb_distance, difference."""
        lat1, lon1 = np.radians(40), np.radians(-74)
        lat2, lon2 = np.radians(51), np.radians(-0.1)

        gc, rhumb, diff = compare_great_circle_rhumb(lat1, lon1, lat2, lon2)

        assert gc > 0
        assert rhumb > 0
        assert diff >= 0  # Rhumb should be >= great circle

    def test_meridional_path_equal(self):
        """Along meridian, both should be equal."""
        lat1, lon1 = np.radians(40), np.radians(-74)
        lat2, lon2 = np.radians(50), np.radians(-74)

        gc, rhumb, diff = compare_great_circle_rhumb(lat1, lon1, lat2, lon2)

        assert diff == pytest.approx(0.0, abs=0.1)

    def test_equatorial_path_equal(self):
        """Along equator, both should be equal."""
        lat1, lon1 = np.radians(0), np.radians(0)
        lat2, lon2 = np.radians(0), np.radians(30)

        gc, rhumb, diff = compare_great_circle_rhumb(lat1, lon1, lat2, lon2)

        assert diff == pytest.approx(0.0, abs=0.1)

    def test_diagonal_path_rhumb_longer(self):
        """Diagonal path should have rhumb > great circle."""
        lat1, lon1 = np.radians(40), np.radians(-74)
        lat2, lon2 = np.radians(51), np.radians(-0.1)

        gc, rhumb, diff = compare_great_circle_rhumb(lat1, lon1, lat2, lon2)

        assert rhumb > gc
        assert diff > 0
